/* (*+ newsstats NEWSSTATS='-prev NSData.NSPrev -in NSData.NSData'*) */
/*
NewsStats
To make the now infamous Posting Statistics list ;-)
RjY, (c) anARCHy, 2000-2001

0.00 (21:2/2/2000) BASIC version
0.10 (23:2/2/2000) C version
     (24:2/2/2000) took it to the unix labs and finished it off.
0.20 (16:2/3/2000) the "generalised" version, handles name and address changes.
0.21 (23:2/3/2000) at most one "From:" line per post
0.30 (18:1/1/2001) Last Week's Position In Brackets
0.31 (21:1/1/2001) finalised new format for table
0.40 (21:3/1/2001) Makes a linked list of posters (not a finite-sized array)
                   "Minimal table width"
0.41 (20:2/2/2001) Poster list self-organises as it is built
0.42 (24:3/2/2001) Output spam-proof email addresses
                   (Thanks to Nick Barton for suggestion)
0.43  (1:4/2/2001) Total posts calculated (with previous in brackets, naturally :-) )
                   Removed cmdline matching options - always "realname or username"
                   (Renamed struct "poster" to "author")
      (2:4/2/2001) Better spam-proof email addresses :-)
      (3:4/2/2001) Update *current* authors with previous
                   Command line case-insensitive ("-in" == "-IN")
0.44  (9:4/2/2001) When sorting: if real names match (ie are blank) try user names
     (23:4/2/2001) When sorting: compare "realname if nonnull, else username"
                   Ignores posts if no realname or username (even if there is a host)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAXINP 1023
char INPUT[MAXINP+1]; /* Global input string buffer, preferable to large auto arrays */

typedef struct {
  char *user, *host, *name;
  int userl, hostl, namel;
} fromline;

typedef struct author_s {
  struct author_s *prev, *next;
  int pos, prevpos, count, prevcount;
  char *user, *host, *name;
} author;

typedef struct {
  author *author;
  int count;
} matches;

author acap = {&acap, &acap, 0, 0, 0, 0, "", "", "TOTAL"};

void nomemory(void)
{
  fprintf(stderr, "newsstats: Not enough memory\n");
}

/* This is now case-insensitive */
int checkparameter(int count, char **args, const char *parm)
{
  int i,j;
  char a,b;

  for (i=1; i<count; i++)
  {
    for (j=0; ; j++)
    {
      a = tolower(parm[j]); b = tolower(args[i][j]);
      if (a != b || a == '\0' || b == '\0') break;
    }
    if (a == b)
      return i;
  }
  return 0;
}

/* if first n chars match and one of the strings has
length n>0, return true(1), else return false(0) */
int comparefromstrs(const char *s1, const char *s2, const int n)
{
  return (n > 0 && (s1[n]=='\0' || s2[n]=='\0') && !strncmp(s1,s2,n));
}

void parsefromline(char *string0, fromline *from)
{
  char *string=string0, *at=NULL, *word0, *word1;
  char *host=NULL, *user=NULL, *name=NULL;
  int hostl=0, userl=0, namel=0;

  if (!strncmp(string0,"From: ",6)) string+=6;

  word0 = string;
  while (1)
  {
    while (*word0 == ' ') word0++;

    for (word1 = word0; *word1 > ' '; word1++)
      if (at == NULL && *word1 == '@') at = word1;

    if (word0 <= at && at < word1)
    {
      user = word0; userl = (at-word0);
      host = at+1;  hostl = (word1-host);
    }
    else
    {
      if (name == NULL) name = word0;
      namel = word1-name;
    }

    if (*word1 < ' ')
      break;
    else
      word0 = word1;
  }

  if (userl > 0 && hostl > 0 && user[0] == '<' && host[hostl-1] == '>')
    { user++; userl--; hostl--; }
  if (namel > 0 && ((name[0] == '('  && name[namel-1] == ')')
   || (name[0] == '\"' && name[namel-1] == '\"')))
    { name++; namel--; namel--; }

  from->user = user; from->host = host; from->name = name;
  from->userl = userl; from->hostl = hostl; from->namel = namel;
}

int addnewauthor(fromline *from, author *newtemplate)
{
  author *new, *tmp;
  int length;
  char *user,*host,*name;

  if (!(new = malloc(sizeof(author)))
    || !(user = malloc(from->userl+1))
    || !(host = malloc(from->hostl+1))
    || !(name = malloc(from->namel+1)))
  {
    nomemory(); return 0;
  }

  *new = *newtemplate;
  acap.count += new->count;

  strncpy(new->user=user, from->user, length=from->userl); new->user[length]='\0';
  strncpy(new->host=host, from->host, length=from->hostl); new->host[length]='\0';
  strncpy(new->name=name, from->name, length=from->namel); new->name[length]='\0';

  tmp = acap.next;
  new->prev = &acap; new->next = tmp;
  acap.next = new; tmp->prev = new;

  return 1;
}

int updateauthor(author *a, fromline *from)
{
  char *user,*host,*name;
  int length;
  author *tmp;

  a->count++;
  acap.count++;

  if (!comparefromstrs(user=a->user, from->user, length=from->userl))
  {
    if (!(user = realloc(user, length+1))) {nomemory(); return 0;}
    strncpy(a->user=user, from->user, length); a->user[length]='\0';
  }
  if (!comparefromstrs(host=a->host, from->host, length=from->hostl))
  {
    if (!(host = realloc(host, length+1))) {nomemory(); return 0;}
    strncpy(a->host=host, from->host, length); a->host[length]='\0';
  }
  if (!comparefromstrs(name=a->name, from->name, length=from->namel))
  {
    if (!(name = realloc(name, length+1))) {nomemory(); return 0;}
    strncpy(a->name=name, from->name, length); a->name[length]='\0';
  }

  if ((tmp = acap.next) != a)
  {
    a->prev->next = a->next; a->next->prev = a->prev;
    a->next = tmp; a->prev = &acap;
    acap.next = a; tmp->prev = a;
  }

  return 1;
}

matches compareposts(fromline *from)
{
  author *a;
  matches matches;

  matches.count = 0; matches.author = &acap;
  for (a = acap.next; a != &acap; a = a->next)
  {
    if ( comparefromstrs(a->user, from->user, from->userl)
      || comparefromstrs(a->name, from->name, from->namel) )
    {
      matches.count++; matches.author = a;
    }
  }
  return matches;
}

int scannews(FILE *nf)
{
  int foundfrom = 0;
  fromline from;
  author new;
  matches matches;

  memset(&new, 0, sizeof(author)); new.count = 1;
  while (1)
  {
    fgets(INPUT,MAXINP,nf);
    if (feof(nf))
      return 1;
    if (!strcmp(INPUT,".\n"))
      foundfrom = 0;
    if (!foundfrom && !strncmp(INPUT,"From: ",6))
    {
      foundfrom = 1;
      parsefromline(INPUT, &from);
      if (!from.userl && !from.namel)
        continue; /* ignore badly formatted From: lines */
      matches = compareposts(&from);
      switch (matches.count)
      {
        case 0:
          if (!addnewauthor(&from, &new)) return 0;
          break;
        case 1:
          if (!updateauthor(matches.author, &from)) return 0;
          break;
      }
    }
  }
}

void handlenews(int argcount, char **args)
{
  int p, check;
  FILE *newsfile;

  if (!(p = checkparameter(argcount, args, "-in")))
  {
    fprintf(stderr, "newsstats: no input news file\n");
    exit(1);
  }
  while ((p < argcount-1) && (args[p+1][0] != '-') )
  {
    if (!(newsfile = fopen(args[p+1], "r")))
    {
      fprintf(stderr, "newsstats: cannot open news file %s\n",args[p+1]);
      exit(1);
    }
    check = scannews(newsfile);
    fclose(newsfile);
    if (!check) exit(1);
    p++;
  }
}

int scanprev(FILE *prf)
{
  int ss, a, b, length, prevtotalcheck = 0;
  fromline from;
  matches matches;

  while (1)
  {
    fgets(INPUT, MAXINP, prf);
    if (feof(prf)) break;
    ss = sscanf(INPUT,"%d %*s %d %*s %n", &a, &b, &length);
    switch (ss)
    {
      case 2: /* Normal entry */
        parsefromline(INPUT + length, &from);
        matches = compareposts(&from);
        if (matches.count == 1)
        {
          matches.author->prevpos = a;
          matches.author->prevcount = b;
        }
        prevtotalcheck += b;
        break;
      case 1: /* Total */
        acap.prevcount = a;
        break;
    }
  }
  /* The first time this gets used we won't have a previous total, so... */
  if (prevtotalcheck != acap.prevcount)
  {
    fprintf(stderr, "newsstats: Warning: Incorrect previous total (read %d, calc %d)\n",
      acap.prevcount, prevtotalcheck);
    acap.prevcount = prevtotalcheck;
  }
  return 1;
}

void handleprev(int argcount, char **args)
{
  int p, check;
  FILE *prevfile;

  p = checkparameter(argcount, args, "-prev");
  if (p && (p < argcount-1))
  {
    if (!(prevfile = fopen(args[p+1], "r")))
    {
      fprintf(stderr, "newsstats: cannot open prev file %s\n",args[p+1]);
      exit(1);
    }
    check = scanprev(prevfile);
    fclose(prevfile);
    if (!check) exit(1);
  }
}

void sortresults(void)
{
  author *a, *b, *tmp;
  int ca, cb, prevca, prevcb, pos;
  char *namea, *nameb;

  for (pos = 1, a = acap.next; a != &acap; pos++, a = a->next)
  {
    for (b = a; b != &acap; b = b->next)
    {
      ca = a->count; cb = b->count; prevca = a->prevcount; prevcb = b->prevcount;
      namea = *a->name ? a->name : a->user; nameb = *b->name ? b->name : b->user;
      if (ca<cb
        || (ca == cb && prevca < prevcb)
        || (ca == cb && prevca == prevcb && (strcmp(namea, nameb)>0)))
      {
        tmp = b->prev;
        b->prev->next = b->next; b->next->prev = b->prev;
        b->next = a; b->prev = a->prev;
        a->prev->next = b; a->prev = b;
        a = b; b = tmp;
      }
    }
    if (pos == 1 || a->count != a->prev->count)
      a->pos = pos;
    else
      a->pos = a->prev->pos;
  }
}

int digits(int n)
{
  int i,c;

  if (n<10) return 1;
  if (n<100) return 2;
  for (i=2, c=100; c<=n; i++, c=10*c);
  return i;
}

void printresults(void)
{
  author *a;
  int maxpos, maxprevpos = 1, maxcount, maxprevcount = 1, fromoffs, ttlspace;
  char format[24], prevpos_b[8], prevcount_b[8], *prevpos, *prevcount;

  for (a = acap.next; a != &acap; a = a->next)
  {
    maxprevpos = a->prevpos > maxprevpos ? a->prevpos : maxprevpos;
    maxprevcount = a->prevcount > maxprevcount ? a->prevcount : maxprevcount;
  }
  maxpos = digits(a->prev->pos); maxprevpos = digits(maxprevpos);
  maxcount = digits(acap.count); maxprevcount = digits(acap.prevcount);

  sprintf(format, "%%%dd %%%ds %%%dd %%%ds ",
    maxpos, maxprevpos+2, maxcount, maxprevcount+2);
  ttlspace = maxpos+1+maxprevpos+2+1;
  fromoffs = ttlspace+maxcount+1+maxprevcount+2+1;

  for (a = acap.next; ; a = a->next)
  {
    if (a->prevpos)
      sprintf(prevpos = prevpos_b, "(%d)", a->prevpos);
    else
      prevpos = "(-)";

    if (a->prevcount)
      sprintf(prevcount = prevcount_b, "(%d)", a->prevcount);
    else
      prevcount = "(-)";

    sprintf(INPUT, format, a->pos, prevpos, a->count, prevcount);

    if (*a->user)
    {
      if (*a->name)
        sprintf(INPUT + fromoffs, "%s <%s@>", a->name, a->user);
      else
        sprintf(INPUT + fromoffs, "%s@", a->user);
    }
    else
      sprintf(INPUT + fromoffs, "%s", a->name);

    if (a != &acap)
      puts(INPUT);
    else
    {
      while (ttlspace > 0) { ttlspace--; INPUT[ttlspace] = ' '; }
      puts(INPUT);
      break;
    }
  }
}

int main(int argcount, char **args)
{
  handlenews(argcount, args);
  handleprev(argcount, args);

  sortresults();
  printresults();

  return 0;
}
